with Messages, Strings, Flts.Std, Protected_Counters;
use Messages, Strings, Flts.Std, Protected_Counters;

pragma Elaborate_All (Messages,Strings,Flts.Std,Protected_Counters);

package body Fouriers2.Contr is
  
  use SSV.SV_Ops;
  
  procedure Show_Degs is
  begin
    Show_Param_F2("Fouriers2 ");
    Show2("Fouriers2.Contr PDegs =",PDeg1,PDeg2);
    Show1("Fouriers2.Contr  VDim =",VDim);
    if not STrunc then
      Show2("Fouriers2.Contr EDegs =",Default_EDeg1,Default_EDeg2);
    end if;
  end Show_Degs;
    
  function NoLowErrs(F: Fourier2) return Boolean is
    E: FErr renames F.E;
  begin
    for M in 0 .. PDeg1 loop
      for N in 0 .. PDeg2 loop
        if E(M,N) /= Zero then return False; end if;
      end loop;
    end loop;
    return True;
  end NoLowErrs;
  
  procedure DistErr(F: in out Fourier2) is
--- distribute errors up to degree (Pdeg1,PDeg2) into coefficients
--- assumes odd F if JustOdd
    W1: constant Flt := NotExp(-F.T.R(1));
    W2: constant Flt := NotExp(-F.T.R(2));
    C: FCoeff renames F.C;
    E: FErr renames F.E;
    EvenM,EvenN: Boolean;
    Fac1,Fac2: Flt;
    B: Scalar;
  begin
    if STrunc or else NoLowErrs(F) then return; end if;
    for M in 0 .. PDeg1+1 loop
      for N in 0 .. PDeg2 loop
        declare
          E1: Flt renames E(M,N+1);
        begin
          E1 := E1+E(M,N);        -- sum along row M
        end;
      end loop;
    end loop;
    EvenN := True;
    Fac2 := One;
    for N in 0 .. PDeg2+1 loop
      EvenM := True;
      Fac1 := Fac2;
      for M in 0 .. PDeg1 loop
        declare
          E0: Flt renames E(M,N);
          E1: Flt renames E(M+1,N);
        begin
          if JustOdd and then (EvenN or else EvenM) then
            SetZero(C(M,N));
          else
            BallAt0(E0*Fac1,B);
            Add(B,C(M,N));
          end if;
          E1 := E1+E0;            -- sum along column N
          E0 := Zero;
        end;
        EvenM := not EvenM;
        Fac1 := W1*Fac1;
      end loop;
      EvenN := not EvenN;
      Fac2 := W2*Fac2;
    end loop;
  end DistErr;
  
  procedure Pack(C: in FCoeff; V: in out PVec) is
    I,J,K: Integer;
  begin
    K := 1;
    J := Lowest;
    while J <= PDeg2 loop
      I := Lowest;
      while I <= PDeg1 loop
        V(K) := C(I,J);
        K := K+1;
        I := I+Increment;
      end loop;
      J := J+Increment;
    end loop;
  end Pack;

  procedure Pack(F: in Fourier2; V: in out PVec) is
    P: LF_Pointer;
  begin
    if (STrunc or else NoLowErrs(F)) then
      Pack(F.C,V);
    else
      Pool.Allocate(P);
      Copy(F,P.Data);
      DistErr(P.Data);
      Pack(P.Data.C,V);
      Pool.Recycle(P);
    end if;
  end Pack;
  
  procedure UnPack(V: in PVec; C: in out FCoeff) is
    I,J,K: Integer;
  begin
    K := 1;
    J := Lowest;
    while J <= PDeg2 loop
      I := Lowest;
      while I <= PDeg1 loop
        C(I,J) := V(K);
        K := K+1;
        I := I+Increment;
      end loop;
      J := J+Increment;
    end loop;
  end UnPack;

  procedure UnPack(V: in PVec; T: in FType; F: in out Fourier2) is
  begin
    SetZero(T,F);
    UnPack(V,F.C);
  end UnPack;

  procedure Prod(A: in PMat; F: in Fourier2; G: in out Fourier2) is
    V: PVec;
  begin
    Pack(F,V);
    Mult(A,V);
    UnPack(V,F.T,G);
  end Prod;
  
  function OpNorm(T: FType; M: PMat) return Radius is
    R: Radius := Zero;
    R0,R1: Radius;
    V: PVec;
    P: LF_Pointer;
  begin
    Pool.Allocate(P);
    for K in V'Range loop
      SetZero(V);
      Assign(1,V(K));
      UnPack(V,T,P.Data);
      R0 := Inf(Norm(P.Data));
      Mult(M,V);
      UnPack(V,T,P.Data);
      R1 := Sup(Norm(P.Data));
      R := RMax(R1/R0,R);
    end loop;
    Pool.Recycle(P);
    return R;
  end OpNorm;
  
  procedure Contract(P: in Param; F: in Fourier2; A: in PMat; H: in out Fourier2; Err: out Radius) is
    P0,P1,P2: LF_Pointer;
  begin
    Pool.Allocate(P0);
    Pool.Allocate(P1);
    Pool.Allocate(P2);
    declare
      AH: Fourier2 renames P0.Data;
      F1: Fourier2 renames P1.Data;
      F2: Fourier2 renames P2.Data;
    begin
      Prod(A,H,AH);
      Sum(H,AH,F1);
      Add(F,F1);
      Map(P,F1,F2);
      Sub(F2,F1);
      Err := MaxNorm(F1);
      Sub(F1,F2);
      Diff(F2,AH,H);
    end;
    Pool.Recycle(P0);
    Pool.Recycle(P1);
    Pool.Recycle(P2);
  end Contract;
  
  function Check_EDeg1(D: Positive) return Positive is
    Min_EDeg1: constant Positive := PDeg1+Increment;
    Max_EDeg1: constant Positive := Deg1+Increment;
    ED1: Integer := IMax(D,Min_EDeg1);
  begin
    if JustOdd and then ((ED1 mod 2)=0) then ED1 := ED1+1; end if;
    return IMin(ED1,Max_EDeg1);
  end Check_EDeg1;
  
  function Check_EDeg2(D: Positive) return Positive is
    Min_EDeg2: constant Positive := PDeg2+Increment;
    Max_EDeg2: constant Positive := Deg2+Increment;
    ED2: Integer := IMax(D,Min_EDeg2);
  begin
    if JustOdd and then ((ED2 mod 2)=0) then ED2 := ED2+1; end if;
    return IMin(ED2,Max_EDeg2);
  end Check_EDeg2;
  
  procedure SetEDegs(EDeg1,EDeg2: in Positive) is
  begin
    Default_EDeg1 := Check_EDeg1(EDeg1);
    Default_EDeg2 := Check_EDeg2(EDeg2);
  end SetEDegs;
  
  procedure Show_Low(ID,I,J: in Integer; R: in Radius) is
  begin
    Show1(Strng(ID,2) & ":  " & Strng(I,4) & Strng(J,4) & " ||low || <=",R);
  end Show_Low;
  
  procedure Show_High(ID,I,J: in Integer; R: in Radius) is
  begin
    Show1(Strng(ID,2) & ":  " & Strng(I,4) & Strng(J,4) & " ||high|| <=",R);
  end Show_High;
  
--------------------------------------------
  
  protected body Protected_Ops is
    
    procedure Initialize(EDeg1,EDeg2: in Positive) is
    begin
      ED1 := EDeg1;
      ED2 := EDeg2;
      I0 := Lowest;
      J0 := Lowest;
      ECount := 2;
      R0 := Zero;
    end Initialize;
    
    procedure Check_Max(ID,I,J: in Integer; R: in Radius) is
    begin
      if Verbosity>1 then
        if (I<ED1) and (J<ED2) then
          Show_Low(ID,I,J,R);
        else
          Show_High(ID,I,J,R);
        end if;
      end if;
      R0 := RMax(R0,R);
    end Check_Max;
    
    procedure Get_Max(R: out Radius) is
    begin
      R := R0;
    end Get_Max;
    
    procedure New_Low_Mode(I,J: out Integer; Done: out Boolean) is
    begin
      I := I0;
      J := J0;
      Done := (I0 >= ED1);
      if not Done then
        J0 := J0+Increment;
        if J0 >= ED2 then
          J0 := Lowest;
          I0 := I0+Increment;
        end if;
      end if;
    end New_Low_Mode;
        
    procedure New_High_Mode(I,J: out Integer; Done: out Boolean) is
    begin
      if ECount=0 then
        Done := True;
      elsif ECount=1 then
        I := ED1;
        J := Lowest;
        Done := False;
        ECount := 0;
      elsif ECount=2 then
        I := Lowest;
        J := ED2;
        Done := False;
        ECount := 1;
      end if;
    end New_High_Mode;
    
  end Protected_Ops;
  
--------------------------------------------
  
  function DContractNorm(P: Param; F: Fourier2; A: PMat; EDeg1,EDeg2: Positive) return Radius is
    Tr: constant Tracer := Starting("Fouriers2.Contr.DContractNorm");
    ED1: constant Positive := Check_EDeg1(EDeg1);
    ED2: constant Positive := Check_EDeg2(EDeg2);
    PC: Protected_Counter;
    
    procedure Mode_Loop(Id: in Integer; H,AH: in out Fourier2) is
      Done: Boolean;
      I,J: Integer;
      R0,R1: Radius;
    begin
      loop
        Prot.New_High_Mode(I,J,Done);
        exit when Done;
        if not STrunc then
          SetZero(F.T,H);
          H.E(I,J) := One;
          DMap(P,F,H);
          R1 := Sup(Norm(H));
          Prot.Check_Max(ID,I,J,R1);
        end if;
      end loop;
      loop
        Prot.New_Low_Mode(I,J,Done);
        exit when Done;
        SetZero(F.T,H);
        Assign(1,H.C(I,J));
        R0 := Inf(Norm(H));
        if (I <= PDeg1) and then (J <= PDeg2) then
          Prod(A,H,AH);
          Add(AH,H);
          DMap(P,F,H);
          Sub(AH,H);
        else
          DMap(P,F,H);
        end if;
        R1 := Sup(Norm(H));
        Prot.Check_Max(ID,I,J,R1/R0);
      end loop;
    end Mode_Loop;
    
    procedure Mode_Loop is --- avoiding STORAGE_ERROR ...
      Id: Integer;
      P,AP: LF_Pointer;
    begin
      PC.Next(Id);
      Pool.Allocate(P);
      Pool.Allocate(AP);
      Mode_Loop(Id,P.Data,AP.Data);
      Pool.Recycle(P);
      Pool.Recycle(AP);
    exception
      when others => Task_Error := True; raise;
    end Mode_Loop;
    
    task type Mode_Task_Type is
      pragma Storage_Size (Stack_Size);
    end Mode_Task_Type;
    task body Mode_Task_Type is begin Mode_Loop; end Mode_Task_Type;
    
    NT: constant Natural := Reserve_Tasks(999,Contr_Parallel);
    R: Radius;
  begin
    PC.Reset(1);
    Prot.Initialize(ED1,ED2);
    if NT=0 then  
      Mode_Loop;
    else
      declare
        Mode_Task: array(1 .. NT) of Mode_Task_Type;
      begin
        Mode_Loop;
      end;
      Free_Tasks(NT);
      if Task_Error then raise Sorry with "Exception in Fouriers2.Contr.DContractNorm"; end if;
    end if;
    Prot.Get_Max(R);
    if Verbosity>1 then
      Show1("R",R);
    end if;
    return R;
  end DContractNorm;
  
  function DContractNorm(P: Param; F: Fourier2; A: PMat) return Radius is
    R: Radius;
  begin
    R := DContractNorm(P,F,A,Default_EDeg1,Default_EDeg2);
    return R;
  end DContractNorm;
  
  procedure CheckFix(P: in Param; F: in out Fourier2; A: in PMat; KMax: in Radius; OK: out Boolean) is
--- assuming Flt operations are rounded up
    Tr: constant Tracer := Starting("Fouriers2.Contr.CheckFix");
    R,Del,N,K: Radius;
    GP: LF_Pointer;
  begin
    Pool.Allocate(GP);
    declare
      G: Fourier2 renames GP.Data;
    begin
      Map(P,F,G);
      Sub(F,G);
      Del := Sup(Norm(G));
      if Verbosity>1 then
        Show1("||Map(F)-F|| <=",Del);
      end if;
      R := (-Del)/(KMax-One);
      N := One+OpNorm(F.T,A);
      Copy(F,G);
      AddBall(R*N,G);
      K := DContractNorm(P,G,A);
      OK := (K <= KMax);
      if OK then
        Show0("CheckFix ok.");
        R := (-Del)/(K-One);    -- radius for contraction
        AddBall(R*N,F);
      else
        Show0("CheckFix failed.");
      end if;
    end;
    Pool.Recycle(GP);
  end CheckFix;
  
  procedure CheckFix(P: in Param; F: in out Fourier2; A: in PMat; KMax: in Radius) is
    OK: Boolean;
  begin
    CheckFix(P,F,A,KMax,OK);
  end CheckFix;
    
  procedure CheckFix(P: in Param; F: in out Fourier2; KMax: in Radius) is
    OK: Boolean;
    A: PMat;
  begin
    ContrMatrix(P,F,A);
    CheckFix(P,F,A,KMax,OK);
  end CheckFix;
  
-------------------------------------- numerical
  
  procedure DerMatrix(P: in Param; F: in Fourier2; A: in out PMat) is
--- use only with numeric scalars
    PC: Protected_Counter;
    
    procedure Column_Loop(H: in out Fourier2) is
      K: Integer;
      V: PVec;
    begin
      loop
        PC.Next(K);
        exit when K>VDim;
        SetZero(V);
        Assign(1,V(K));
        UnPack(V,F.T,H);
        DMap(P,F,H);
        Pack(H,V);
        SetColumn(K,V,A);
      end loop;
    end Column_Loop;
    
    procedure Column_Loop is
      P: LF_Pointer;
    begin
      Pool.Allocate(P);
      Column_Loop(P.Data);
      Pool.Recycle(P);
    end Column_Loop;
    
    task type Column_Task_Type is
      pragma Storage_Size (Stack_Size);
    end Column_Task_Type;
    task body Column_Task_Type is begin Column_Loop; end Column_Task_Type;
    
    NT: constant Natural := Reserve_Tasks(VDim-1,Contr_Parallel);
  begin
    PC.Reset(1);
    if NT=0 then  
      Column_Loop;
    else
      declare
        Column_Task: array(1 .. NT) of Column_Task_Type;
      begin
        Column_Loop;
      end;
      Free_Tasks(NT);
    end if;
  end DerMatrix;
    
  procedure ContrMatrix(P: in Param; F: in Fourier2; A: in out PMat) is
--- use only with numeric scalars
  begin
    DerMatrix(P,F,A);
    AddConst(-Scal(One),A);
    Invert(A);              -- inv(DG-I)
    AddConst(Scal(One),A);  -- inv(DG-I)+I
    Neg(A);                 -- -I-inv(DG-I)
  end ContrMatrix;
  
  procedure IterContr(P: in Param; F: in out Fourier2; A: in out PMat; Eps: in Radius) is
--- use only with numeric scalars
    MaxErr: constant Radius := 1.0E+9;
    Count: Natural := 100;
    LastErr: Radius := MaxErr;
    
    function Check(Err: Radius) return Boolean is
    begin
      if Err<Eps then
        if Err >= LastErr then return True; end if;
      else
        Count := Count-1;
        if Count=0 or Err>MaxErr then return True; end if;
      end if;
      LastErr := Err;
      return False;
    end Check;
    
    Err: Radius;
    P0,P1,P2: LF_Pointer;
  begin
    Pool.Allocate(P0);
    Pool.Allocate(P1);
    Pool.Allocate(P2);
    declare
      F0: Fourier2 renames P0.Data;
      H:  Fourier2 renames P1.Data;
      AH: Fourier2 renames P2.Data;
    begin
      if Verbosity>1 then Show0("IterContr errors:"); end if;
      for I in 1 .. Integer'Last loop --- contract0
        Copy(F,F0);
        Map(P,F0,F);
        Diff(F,F0,H);
        Err := MaxNorm(H);
        if Verbosity>1 then Show1(Strng(I,3),Err); end if;
        Prod(A,H,AH);
        Add(AH,F);
        exit when Check(Err);
      end loop;
    end;
    Pool.Recycle(P0);
    Pool.Recycle(P1);
    Pool.Recycle(P2);
    if Err<Eps then
      ContrMatrix(P,F,A);
    else
      Show0("IterContr: getting tired");
    end if;
  end IterContr;
  
  procedure ApproxFix(P: in Param; F: in out Fourier2; A: in out PMat; Eps: in Radius) is
--- use only with numeric scalars
    MaxErr: constant Radius := 1.0E+9;
    Count: Natural := 100;
    LastErr: Radius := MaxErr;
    
    function Check(Err: Radius) return Boolean is
    begin
      if Err<Eps then
        if Err >= LastErr then return True; end if;
      else
        Count := Count-1;
        if Count=0 or Err>MaxErr then return True; end if;
      end if;
      LastErr := Err;
      return False;
    end Check;
    
    Err: Radius;
    P0,P1,P2: LF_Pointer;
  begin
    Pool.Allocate(P0);
    Pool.Allocate(P1);
    Pool.Allocate(P2);
    declare
      F0: Fourier2 renames P0.Data;
      H:  Fourier2 renames P1.Data;
      AH: Fourier2 renames P2.Data;
    begin
      if Verbosity>1 then Show0("ApproxFix errors:"); end if;
      for I in 1 .. Integer'Last loop --- contract0
        ContrMatrix(P,F,A);
        Copy(F,F0);
        Map(P,F0,F);
        Diff(F,F0,H);
        Err := MaxNorm(H);
        if Verbosity>1 then Show1(Strng(I,3),Err); end if;
        Prod(A,H,AH);
        Add(AH,F);
        exit when Check(Err);
      end loop;
    end;
    Pool.Recycle(P0);
    Pool.Recycle(P1);
    Pool.Recycle(P2);
    if (Verbosity>1) and then (Err>Eps) then
      Show0("ApproxFix: getting tired");
    end if;
  end ApproxFix;
  
  procedure ApproxFix(P: in Param; F: in out Fourier2; Eps: in Radius) is
--- use only with numeric scalars
    A: PMat;
  begin
    ApproxFix(P,F,A,Eps);
  end ApproxFix;
  
  procedure Eigen(P: in Param; F: in Fourier2; Wr: in out PVec; Large: out Integer) is
--- use only with numeric scalars    
    use SSV_RNum;
    Wi: PVec;
    A,Vr,Vi: PMat;
  begin
    DerMatrix(P,F,A);
    Eigen(A,Wr,Wi,Vr,Vi);
--    for K in 1 .. 10 loop
--      Show2(Strng(K) & ": ",Wr(K),Wi(K));
--    end loop;
    Large := 0;
    for K in Wr'Range loop
      if MaxNorm(Wr(K))>One then Large := Large+1; end if;
    end loop;
  end Eigen;
  
  procedure Eigen(P: in Param; F: in Fourier2; Large: out Natural) is
--- use only with numeric scalars
    Wr: PVec;
  begin
    Eigen(P,F,Wr,Large);
  end Eigen;
  
  procedure NumRead(Name: in String; A: in out PMat; Decimal: in Boolean := False) is
  begin
    Read(Name,A,Decimal);
    Center(A);
  end NumRead;
  
end Fouriers2.Contr;
