with Strings, Globals, Protected_Counters, Flt_Vectors;
use Strings, Globals, Protected_Counters;

pragma Elaborate_All(Strings,Globals,Protected_Counters,Flt_Vectors);

package body MultiTaylors1.Phi is

  use SM,SV,STM,STV;
  
  procedure Show_Param is
  begin
    Show_Param_T1("S");
    Show1("M1.Phi DimX =",DimX);
    Show1("M1.Phi FreqParity =",FreqParity);
  end Show_Param;
  
  function Symplecticity(U: Scalar_Matrix) return Radius is
    -- a norm of transposed(U)*J*U-J
    -- using formula (1.2) in the notes
    L: constant Positive := U'Last(1)/2;
    NT: constant Natural := Reserve_Tasks(L-1,Flow_Parallel);
    PC: Protected_Counter;
    E: Flt_Vector(1 .. L);
    
    procedure Symp_Loop is
      I1,I2,J2,K2: Integer;
      AD,CB,AC,CA,DB,BD,E1,E2,E3,Tmp: Scalar;
    begin
      loop
        PC.Next(I1);
        exit when I1>L;
        I2 := I1+L;
        for J1 in 1 .. L loop
          J2 := J1+L;
          SetZero(AD); SetZero(CB);
          SetZero(AC); SetZero(CA);
          SetZero(DB); SetZero(BD);
          for K1 in 1 .. L loop
            K2 := K1+L;
            AddProd(U(K1,I1),U(K2,J2),AD,Tmp);
            AddProd(U(K2,I1),U(K1,J2),CB,Tmp);
            AddProd(U(K1,I1),U(K2,J1),AC,Tmp);
            AddProd(U(K2,I1),U(K1,J1),CA,Tmp);
            AddProd(U(K2,I2),U(K1,J2),DB,Tmp);
            AddProd(U(K1,I2),U(K2,J2),BD,Tmp);
          end loop;
          Diff(AD,CB,E1); -- transposed(A)*D-transposed(C)*B = I
          Diff(AC,CA,E2); -- transposed(A)*C-transposed(C)*A = 0
          Diff(DB,BD,E3); -- transposed(D)*B-transposed(B)*D = 0
          if J1=I1 then Sub(SOne,E1); end if;
          E(I1) := E(I1)+MaxNorm(E1)+MaxNorm(E2)+MaxNorm(E3);
        end loop;
      end loop;
    end Symp_Loop;
    
    task type Symp_Task_Type is end Symp_Task_Type;
    task body Symp_Task_Type is begin Symp_Loop; end Symp_Task_Type;
    
    use Flt_Vectors;
    Err: Radius := Zero;
  begin
    SetZero(E);
    PC.Reset(1);
    if NT=0 then
      Symp_Loop;
    else
      declare
        Symp_Task: array(1 .. NT) of Symp_Task_Type;
      begin
        Symp_Loop;
      end;
      Free_Tasks(NT);
    end if;
    for I in 1 .. L loop Err := Err+E(I); end loop;
    return Err;
  end Symplecticity;

  -------------------- for LinFlow

  function MaxNorm(V: SVector) return Radius is
    --- upper bound on the norm
    S: Radius := Zero;
  begin
    for I in reverse V'Range loop
      S := S+MaxNorm(V(I));
    end loop;
    return S;
  end MaxNorm;

  procedure AddBall(R: in Radius; V: in out SVector) is
    S: Scalar;
  begin
    BallAt0(R,S);
    for I in V'Range loop Add(S,V(I)); end loop;
  end AddBall;

  procedure Quot(V: in SVector; R: in Flt; W: in out SVector) is
  begin
    for I in V'Range loop Quot(V(I),R,W(I)); end loop;
  end Quot;

  function OpNorm(A: SMatrix) return Radius is
    --- upper bound on operator norm
    S: Radius := Zero;
    SJ: Radius;
  begin
    for J in reverse A'Range(2) loop
      SJ := Zero;
      for I in reverse A'Range(1) loop
        SJ := SJ+MaxNorm(A(I,J));
      end loop;
      if SJ>S then S := SJ; end if;
    end loop;
    return S;
  end OpNorm;

  procedure LinFlow(X: in TMatrix; V: in out TVector; DT: in Radius; Tmp1,Tmp2: in out SVector) is
    use SSV_Ops;
    --- solves V'(t)=X(t)V(t) with given V(0)
    --- coefficient-errors can depend on t
    E,Contr: Flt;
    S: SVector renames Tmp1;
  begin
    Prod(X(0),V(0),V(1));
    for M in 1 .. PDeg-1 loop   -- compute V(M+1)
      Prod(X(0),V(M),S);
      for I in 1 .. M loop
        Prod(X(I),V(M-I),Tmp2);
        Add(Tmp2,S);
      end loop;
      Quot(S,Flt(M+1),V(M+1));
    end loop;
    if not STrunc then
      E := Zero;
      for M in reverse PDeg .. 2*PDeg loop
        Prod(X(M-PDeg),V(PDeg),S);
        for I in M-PDeg+1 .. PDeg loop
          Prod(X(I),V(M-I),Tmp2);
          Add(Tmp2,S);
        end loop;
        E := DT*E+MaxNorm(S)/Flt(M+1);
      end loop;                         -- E is degree PDeg+1 (or higher) error
      Contr := Zero;
      for M in reverse 0 .. PDeg loop
        Contr := DT*Contr+OpNorm(X(M));
      end loop;
      Contr := Contr*DT/Flt(PDeg+2);    -- contraction factor at degree D+1 (or higher)
      if Contr>=One then raise Sorry with "LinFlow: no contraction"; end if;
      E := E/abs(Contr-One);            -- assuming Flt is rounded up
      AddBall(DT*E,V(PDeg));
    end if;
  end LinFlow;

  procedure LinFlow(X: in TMatrix; V: in out TVecVec; DT: in Radius) is
    NT: constant Natural := Reserve_Tasks(Dim2X-1,Flow_Parallel);
    Task_Error: Boolean := False;
    PC: Protected_Counter;

    procedure Flow_Loop is
      I: Natural;
      Tmp1,Tmp2: LT_Pointer;
    begin
      Pool.Allocate(Tmp1);
      Pool.Allocate(Tmp2);
      loop
        PC.Next(I);
        exit when I>Dim2X;
        LinFlow(X,V(I),DT,Tmp1.Data,Tmp2.Data);
      end loop;
      Pool.Recycle(Tmp1);
      Pool.Recycle(Tmp2);
    end Flow_Loop;

    task type Flow_Task_Type is end Flow_Task_Type;
    task body Flow_Task_Type is begin
      Flow_Loop;
    exception
      when others => Task_Error := True; raise;
    end Flow_Task_Type;

  begin
    PC.Reset(1);
    if NT=0 then
      Flow_Loop;
    else
      declare
        Flow_Task: array(1 .. NT) of Flow_Task_Type;
      begin
        Flow_Loop;
      end;
      Free_Tasks(NT);
      if Task_Error then raise Sorry with "Exception in MultiTaylors1.Phi.LinFlow"; end if;
    end if;
  end LinFlow;

  ------------------------
  
  procedure SetZero(V: in out TVector) is
  begin
    for M in Power loop SetZero(V(M)); end loop;
  end SetZero;
  
  procedure T2T(A: in out TMatrix1; B: in out TMatrix) is
  begin
    for I in Index2X loop
      for J in Index2X loop
        declare
          AC: Poly1 renames A(I,J).C;
        begin
          for M in Power loop Swap(AC(M),B(M)(I,J)); end loop;
        end;
      end loop;
    end loop;
  end T2T;
  
  procedure HTilde(M: in Scalar_Matrix; Time,DT: in Radius; Y: in Scalar_Vector; H: in out TMatrix1) is
    STime: constant Scalar := Scal(Time);
    I2,Dif,Sum: Integer;
    S,Tmp: Scalar;
    T: Taylor1;
  begin
    SetZero(H);
    for N in M'Range(1) loop
      CosTaylor(N,STime,DT,T);
      for I in IndexX loop
        I2 := I+DimX;
        for J in IndexX loop
          Dif := 2*abs(I-J);
	  Sum := 2*(KOffset+I+J);
          if Dif=0 then
            Prod(STwo,M(N,Dif),S);
            Sub(M(N,Sum),S);
          else
            Diff(M(N,Dif),M(N,Sum),S);
          end if;
          AddProd(S,T,H(I2,J));
        end loop;
      end loop;
    end loop;
    for I in IndexX loop
      Prod(NegHalf,Y(I),S);
      I2 := I+DimX;
      for J in IndexX loop
        Prod(S,Y(J),Tmp);
        Mult(Tmp,H(I2,J)); -- mult by -Y(I)*Y(J)/2
      end loop;
    end loop;
  end HTilde;

  procedure MakeY(I: in Positive; Beta2,C: in Scalar; Y: in out Scalar) is
    -- Y = Sqrt(K*K*Beta^2+C), where
    K: constant Integer := KOffset+2*I;
    S: Scalar;
  begin
    Prod(Flt(K*K),Beta2,S);
    Add(C,S);
    Sqrt(S,Y);
  end MakeY;
  
  function MakeYi(Beta2: Scalar; C: Scalar_Vector) return YType is
    Y,S: Scalar;
    Yi: YType;
  begin
    for I in IndexX loop
      MakeY(I,Beta2,C(I),Y);
      Inv(Y,S);
      Sqrt(S,Yi(I));
    end loop;
    return Yi;
  end MakeYi;

  function MakePhi0(DT: Radius; Beta2: Scalar; C: Scalar_Vector) return TMatrix1 is
    I2: Integer;
    Y: Scalar;
    P: TMatrix1;
  begin
    SetZero(P);
    for I1 in IndexX loop
      I2 := I1+DimX;
      MakeY(I1,Beta2,C(I1),Y);
      CosTaylor(Y,SZero,DT,P(I1,I1));
      Copy(P(I1,I1),P(I2,I2));
      SinTaylor(Y,SZero,DT,P(I1,I2));
      Neg(P(I1,I2),P(I2,I1));
    end loop;
    return P;
  end MakePhi0;
  
  procedure ValPhi0(T,Beta2: in Scalar; C: in Scalar_Vector; V: in out SMatrix) is
    I2: Integer;
    Y: Scalar;
  begin
    SetZero(V);
    for I1 in IndexX loop
      I2 := I1+DimX;
      MakeY(I1,Beta2,C(I1),Y);
      Mult(T,Y);
      Cos(Y,V(I1,I1));
      Copy(V(I1,I1),V(I2,I2));
      Sin(Y,V(I1,I2));
      Neg(V(I1,I2),V(I2,I1));
    end loop;
  end ValPhi0;
  
  procedure Val(P: in TVector; X: in Scalar; Y: in out SVector) is
--- assuming X in the domain
    use SSV_Ops;  
  begin
    Copy(P(PDeg),Y);
    for I in reverse 0 .. PDeg-1 loop
      Mult(X,Y);
      Add(P(I),Y);
    end loop;
  end Val;
  
  procedure Run(DT: in Radius; M: in Scalar_Matrix; Beta2: in Scalar; C: in Scalar_Vector; Phi00: in out SMatrix) is
    TwoPI: constant Scalar := Two*Pi;
    ScalDT: constant Scalar := Scal(DT);
    Yi: constant YType := MakeYi(Beta2,C);
    Ph0: constant TMatrix1 := MakePhi0(DT,Beta2,C);
    Ph0t: constant TMatrix1 := Transpose(Ph0);
    Time: Radius := Zero;
    T0: Scalar := SZero;
    U: SVector;
    ValPh0: SMatrix;
    H,XX,Tmp: TMatrix1;
    X: TMatrix;
    PhH: TVecVec;
  
    procedure MultVal(T: in Scalar) is
      use SSV_Ops;
    begin
      if not (T=T0) then
        ValPhi0(T,Beta2,C,ValPh0);
        Copy(T,T0);
      end if;      
      for I in Index2X loop
        Val(PhH(I),T,U);
        Mult(ValPh0,U);
        SetZero(PhH(I));
        Copy(U,PhH(I)(0));
      end loop;
    end MultVal;
    
    procedure Update_Phi00 is
      use SSV;
    begin
      for I in Index2X loop
	SetColumn(I,PhH(I)(0),Phi00);
      end loop;
      if Verbosity>0 and then Flt'Floor(Time)=Time then
        Show1("Time =",Time,False);
        Show1("  Symplectic Error =",Symplecticity(Phi00));
      end if;
    end Update_Phi00;
    
  begin
    Use_FBalls := True;
    STMProd_Parallel := Flow_Parallel;
    STVProd_Parallel := Flow_Parallel;
    if Verbosity>0 then Show1("DT =",DT); end if;
    for I in Index2X loop
      SetZero(PhH(I));
      Copy(SOne,PhH(I)(0)(I));
    end loop;
    loop
      HTilde(M,Time,DT,Yi,H);
      Prod(H,Ph0,Tmp);
      Prod(Ph0t,Tmp,XX);
      T2T(XX,X);
      LinFlow(X,PhH,DT);
      if Scal(Time+DT)<TwoPI then
        MultVal(ScalDT);                          -- updates PhH
        Time := Time+DT;                          -- chosen to be representble
        if Verbosity>1 then Update_Phi00; end if; -- just for info
      else
        MultVal(TwoPI-Scal(Time));                -- updates PhH
        Time := Approx(TwoPI);                    -- just for info
        exit;
      end if;
    end loop;
    Update_Phi00;
  end Run;
  
  procedure Free_Cache_MT1 is
  begin
    Pool.Free_All;
    Free_Cache_T1;
  end Free_Cache_MT1;
  
end MultiTaylors1.Phi;
