with Protected_Counters;
use Protected_Counters;

pragma Elaborate_All (Protected_Counters);

package body Vectors.Ops is
  
  procedure Mult(S: in Scalar; V: in out Vector) is
  begin
    for N in V'First .. EffLast(V) loop
      Mult(S,V(N));
    end loop;
  end Mult;

  procedure Prod(S: in Scalar; V1: in Vector; V2: in out Vector) is
  begin
    if CheckRange and then ((V1'First /= V2'First) or (V1'Last /= V2'Last)) then
      raise Dimension_Error with "Vectors.Prod error";
    end if;
    for N in V2'Range loop
      Prod(S,V1(N),V2(N));
    end loop;
  end Prod;

  function "*"(S: Scalar; V: Vector) return Vector is
    W: Vector(V'Range);
  begin
    for N in V'Range loop
      Prod(S,V(N),W(N));
    end loop;
    return W;
  end "*";
  
  procedure AddProd(S: in Scalar; V1: in Vector; V2: in out Vector) is
  begin
    if IsZero(S) then return; end if;
    if CheckRange and then ((V1'First /= V2'First) or (V1'Last /= V2'Last)) then
      raise Dimension_Error with "Vectors.AddProd error";
    end if;
    for N in V1'First .. EffLast(V1) loop
      AddProd(S,V1(N),V2(N));
    end loop;
  end AddProd;
  
  procedure Mult(A: in Matrix; V,Tmp: in out Vector) is
  begin
    Copy(V,Tmp);
    Prod(A,Tmp,V);
  end Mult;

  procedure Mult(A: in Matrix; V: in out Vector) is
  begin
    if not IsZero(V) then
      declare
        Tmp: Vector(V'Range);
      begin
        Copy(V,Tmp);
        Prod(A,Tmp,V);
      end;
    end if;
  end Mult;
    
  procedure Prod(A: in Matrix; V1: in Vector; V2: in out Vector) is
    NT: constant Natural := Reserve_Tasks(V2'Last-V2'First,VProd_Parallel);
    F: constant Integer := V1'First;
    L: constant Integer := V1'Last;
  begin
    if CheckRange and then ((A'First(1) /= V2'First) or (A'Last(1) /= V2'Last)
                         or (A'First(2) /= V1'First) or (A'Last(2) /= V1'Last)) then
      raise Dimension_Error with "Vectors.Prod error";
    end if;
    if NT>0 then
      SetZero(V2);
      AddProd_Parallel(A,V1,V2,NT);
      Free_Tasks(NT);
    else
      for I in V2'Range loop
        declare
          V2I: Component renames V2(I);
        begin
          Prod(A(I,F),V1(F),V2I);
          for J in F+1 .. L loop
            AddProd(A(I,J),V1(J),V2I);
          end loop;
        end;
      end loop;
    end if;
  end Prod;
  
  procedure TransProd(A: in Matrix; V1: in Vector; V2: in out Vector) is
    F: constant Integer := V1'First;
    L: constant Integer := V1'Last;
  begin
    if CheckRange and then ((A'First(2) /= V2'First) or (A'Last(2) /= V2'Last)
                         or (A'First(1) /= V1'First) or (A'Last(1) /= V1'Last)) then
      raise Dimension_Error with "Vectors.TransProd error";
    end if;
    for I in V2'Range loop
      declare
        V2I: Component renames V2(I);
      begin
        Prod(A(F,I),V1(F),V2I);
        for J in F+1 .. L loop
          AddProd(A(J,I),V1(J),V2I);
        end loop;
      end;
    end loop;
  end TransProd;

  function "*"(A: Matrix; V: Vector) return Vector is
    W: Vector(A'Range(1));
  begin
    Prod(A,V,W);
    return W;
  end "*";

  procedure AddProd(A: in Matrix; V1: in Vector; V2: in out Vector) is
    NT: constant Natural := Reserve_Tasks(V2'Last-V2'First,VProd_Parallel);
  begin
    if CheckRange and then ((A'First(1) /= V2'First) or (A'Last(1) /= V2'Last)
                         or (A'First(2) /= V1'First) or (A'Last(2) /= V1'Last)) then
      raise Dimension_Error with "Vectors.AddProd error";
    end if;
    if NT>0 then
      AddProd_Parallel(A,V1,V2,NT);
      Free_Tasks(NT);
    else
      for I in V2'Range loop
        declare
          V2I: Component renames V2(I);
        begin
          for J in V1'Range loop
            AddProd(A(I,J),V1(J),V2I);
          end loop;
        end;
      end loop;
    end if;
  end AddProd;
  
  procedure AddProd_Parallel(A: in Matrix; V1: in Vector; V2: in out Vector; NT: in Positive) is
    F: constant Integer := V2'First;
    L: constant Integer := V2'Last;
    PC: Protected_Counter;
    
    procedure Prod_Loop is
      I: Integer;
    begin
      loop
        PC.Next(I);
        exit when I>L;
        for J in V1'Range loop
          AddProd(A(I,J),V1(J),V2(I));
        end loop;
      end loop;
    exception
      when others => Task_Error := True; raise;
    end Prod_Loop;
    
    task type Prod_Task_Type is end Prod_Task_Type;
    task body Prod_Task_Type is begin Prod_Loop; end Prod_Task_Type;
  begin
    PC.Reset(F);
    declare
      Prod_Task: array(1 .. NT) of Prod_Task_Type;
    begin
      Prod_Loop;
    end;
    if Task_Error then raise Sorry with "Exception in Vectors.AddProd_Parallel"; end if;
  end AddProd_Parallel;
  
----------------- Polynom1
  
  procedure Evaluate(P: in Polynom1; S: in Scalar; C: in out Component; Deg: in Natural) is
  begin
    if Deg=0 or else IsZero(S) then
      Copy(P(0),C);
    else
      Copy(P(Deg),C);
      for I in reverse 0 .. Deg-1 loop
        Mult(S,C);
        Add(P(I),C);
      end loop;
    end if;
  end Evaluate;
  
  procedure Evaluate(P: in Polynom1; S: in Scalar; C: in out Component) is
  begin
    Evaluate(P,S,C,EffLast(P));
  end Evaluate;
  
  function Evaluate(P: Polynom1; S: Scalar; Deg: in Natural) return Component is
    C: Component;
  begin
    Evaluate(P,S,C,Deg);
    return C;
  end Evaluate;
  
  function Evaluate(P: Polynom1; S: Scalar) return Component is
    C: Component;
  begin
    Evaluate(P,S,C,EffLast(P));
    return C;
  end Evaluate;
  
  procedure Evaluate(P: in Polynom1; Sr,Si: in Scalar; Cr,Ci: in out Component) is
--- P is assumed to be real
    Deg: constant Natural := EffLast(P);  
  begin
    SetZero(Ci);
    if IsZero(Si) then
      Evaluate(P,Sr,Cr,Deg);
    else
      Copy(P(Deg),Cr);
      declare
        Ctr,Cti: Component;
      begin
        for I in reverse 0 .. Deg-1 loop
          Copy(Cr,Ctr);
          Prod(Si,Ci,Cr);
          Neg(Cr);
          AddProd(Sr,Ctr,Cr); -- Cr := Sr*Cr-Si*Ci
          Copy(Ci,Cti);
          Prod(Si,Ctr,Ci);
          AddProd(Sr,Cti,Ci); -- Ci := Sr*Ci+Si*Cr
          Add(P(I),Cr);
        end loop;
      end;
    end if;
  end Evaluate;
  
  procedure Evaluate(P: in Polynom2; S1,S2: in Scalar; C: in out Component) is
--- assuming P includes all coefficients for degree Deg
    Deg: constant Natural := EffDeg(P);
    C2: Component;
  begin
    if Deg=0 then
      Copy(P(0),C);
    else
      SetZero(C);
      for I in reverse 0 .. Deg loop
        SetZero(C2);
        for D in reverse I .. Deg loop -- J=D-I
          Mult(S2,C2);
          Add(P((D*(D+3))/2-I),C2);
        end loop;
        Mult(S1,C);
        Add(C2,C);
      end loop;
    end if;
  end Evaluate;
  
  procedure Translate(P1: in Polynom1; S: in Scalar; P2: in out Polynom1) is
    D1: constant Natural := EffDeg(P1);
    X: array(0..D1) of Scalar;
  begin
    Copy(P1,P2);
    Assign(1,X(0));
    for N in 1 .. D1 loop
      Assign(1,X(N));
      for K in reverse 1 .. N-1 loop  -- produce X(t)=(t+S)^N
        Mult(S,X(K));
        Add(X(K-1),X(K));
        AddProd(X(K),P1(N),P2(K));    -- and add P1(N)*X to P2
      end loop;
      Mult(S,X(0));
      AddProd(X(0),P1(N),P2(0));
    end loop;
  end Translate;
  
  procedure Mult1(S: in Scalar; P: in out Polynom1) is
--- assuming P'First=0
    I: Integer := P'Last;
  begin
    while IsZero(P(I)) loop I := I-1; end loop;
    Neg(P(I),P(I+1));
    while I>0 loop
      Mult(S,P(I));
      Sub(P(I-1),P(I));
      I := I-1;
    end loop;
    Mult(S,P(0));
  end Mult1;
      
  procedure Prod1(S: in Scalar; P1: in Polynom1; P2: in out Polynom1) is
  begin
    Copy(P1,P2);
    Mult1(S,P2);
  end Prod1;
  
end Vectors.Ops;
